/*
 *  canard_search.cpp
 */

#include <iostream>
#include <sstream>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define WANT_OBFUSCATING_OPERATORS
#include <cln/cln.h>

#include <sundials_types.h>
#include <sundials_dense.h>
#include <cvode.h>
#include <nvector_serial.h>
#include <cvode_dense.h>

#include "relaxosc_clncv_proto.h"

using namespace std;

#define ZERO   RCONST(0.0)
#define ONE    RCONST(1.0)
#define TWO    RCONST(2.0)
#define NEGTWO RCONST(-2.0)

#define DIGITS 100

int assign(char *str[], int ns, realtype v[], char *a)
    {
    int i;
    char name[256];
    char *e;

    e = strchr(a,'=');
    if (e == NULL)
        return(-1);
    *e = '\0';
    strcpy(name,a);
    *e = '=';
    ++e;
    for (i = 0; i < ns; ++i)
        if (strcmp(str[i],name)==0)
            break;
    if (i == ns)
        return(-1);
    string estr;
    stringstream ss;
    ss << e << "_" << DIGITS;
    ss >> estr;
    v[i] = estr.c_str();
    return(i);
    }


int RootFunc(realtype t, N_Vector y_, realtype *gout, void *g_data)
    {
    realtype x = NV_Ith_S(y_,0);
    gout[0] = x+ONE;
    gout[1] = x-ONE;
    return 0;
    }

int main (int argc, char *argv[])
    {
    set_realtype_precision(DIGITS);

    int i,j;
    int flag;
    const int N_ = 2;
    const int P_ = 2;
    realtype def_y_[2] = {RCONST(1.0), RCONST(0.0)};
    realtype y_[2];
    realtype def_p_[2] = {RCONST(5.0000000000000003e-02), RCONST(0.0000000000000000e+00)};
    realtype p_[2];
    realtype solver_param_[3] = {"1.0e-16_100","1.0e-14_100",RCONST(10.0)};
    char *varnames_[2] = {"x", "y"};
    char *parnames_[2] = {"epsilon", "a"};
    char *solver_param_names_[3] = {"abserr","relerr","stoptime"};

    for (i = 0; i < N_; ++i)
        y_[i] = def_y_[i];
    for (i = 0; i < P_; ++i)
        p_[i] = def_p_[i];
    for (i = 1; i < argc; ++i)
        {
        int j;
        if (strcmp(argv[i],"-h") == 0)
            {
            cout << "Sorry, the -h option is temporarily disabled.\n";
            // use(argc,argv,N_,varnames_,def_y_,P_,parnames_,def_p_);
            exit(0);
            }
        j = assign(varnames_,N_,y_,argv[i]);
        if (j == -1)
            {
            j = assign(parnames_,P_,p_,argv[i]);
            if (j == -1)
                {
                j = assign(solver_param_names_,3,solver_param_,argv[i]);
                if (j == -1)
                    {
                    fprintf(stderr,"unknown argument: %s\n",argv[i]);
                    // use(argc,argv,N_,varnames_,def_y_,P_,parnames_,def_p_); 
                    exit(-1);
                    }
                }
            }
        }

    N_Vector y0_;
    y0_ = N_VNew_Serial(N_);
    for (i = 0; i < N_; ++i)
        NV_Ith_S(y0_,i) = y_[i];

    void *cvode_mem;
    // For non-stiff problems:
    // cvode_mem = CVodeCreate(CV_ADAMS,CV_FUNCTIONAL);
    // For stiff problems:
    cvode_mem = CVodeCreate(CV_BDF,CV_NEWTON);

    realtype t = ZERO;
    flag = CVodeMalloc(cvode_mem,relaxosc_vf,t,y0_,CV_SS,solver_param_[1],&(solver_param_[0]));
    flag = CVodeSetFdata(cvode_mem,&(p_[0]));
    flag = CVDense(cvode_mem,N_);
    flag = CVDenseSetJacFn(cvode_mem,relaxosc_jac,&(p_[0]));

    flag = CVodeRootInit(cvode_mem,2,RootFunc,&(p_[0]));

    CVodeSetMaxNumSteps(cvode_mem,5000);  // A value much bigger than the default 500.

    realtype t1 = solver_param_[2];
    int nout = 100001;
    realtype h = t1/cl_float(nout-1,cln_precision_);

    cout.precision(10);
    cout << scientific;

    p_[0] = "0.005_100"; // epsilon
    p_[1] = "-1.5_100";  // a

    realtype yl = RCONST(-0.6999); // Jumps left
    realtype yr = RCONST(-0.7000); // Jumps right

    for (int k = 0; k < 250; ++k)
        {
        cout << "L: " << yl << endl;
        cout << "R: " << yr << endl;
        t = ZERO;
        realtype ymid = (yl+yr)/TWO;
        NV_Ith_S(y0_,0) = NEGTWO;
        NV_Ith_S(y0_,1) = ymid;
        flag = CVodeReInit(cvode_mem, relaxosc_vf, t, y0_, CV_SS, solver_param_[1], &(solver_param_[0]));
        if (flag != CV_SUCCESS)
            {
            fprintf(stderr,"CVodeReInit failed\n");
            break;
            }

        int firstroot = 0;
        for (i = 1; i < nout; ++i)
            {
            t = cl_float(i,cln_precision_)*h;
            // Advance the solution
            flag = CVode(cvode_mem,t,y0_,&t,CV_NORMAL);
            if (flag == CV_ROOT_RETURN)
                {
                int rootindex[2];
                CVodeGetRootInfo(cvode_mem,rootindex);
                if (rootindex[0] == 1 & firstroot==0)
                    firstroot = 1;
                else
                    break;
                }
            }
        if (NV_Ith_S(y0_,0) > ZERO)
            {
            yr = ymid;
            }
        else
            {
            yl = ymid;
            }
        }
    N_VDestroy_Serial(y0_);
    CVodeFree(&cvode_mem);
    }
